from .opt_output import OptOutput as OptOutput

from .results_writer import SingleObjResultsWriter as SingleObjResultsWriter
from .results_writer import MultiObjResultsWriter as MultiObjResultsWriter



"""
Wind farm optimization objectives.
"""

from .farm_vars import FarmVarObjective as FarmVarObjective
from .farm_vars import MaxFarmPower as MaxFarmPower
from .farm_vars import MinimalMaxTI as MinimalMaxTI

from .max_n_turbines import MaxNTurbines as MaxNTurbines



import numpy as np
import argparse
import matplotlib.pyplot as plt
from iwopy.interfaces.pymoo import Optimizer_pymoo

import foxes
from foxes_opt.problems.layout import FarmLayoutOptProblem
from foxes_opt.constraints import FarmBoundaryConstraint, MinDistConstraint
from foxes_opt.objectives import MaxFarmPower

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-nt", "--n_t", help="The number of turbines", type=int, default=10
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Bastankhah025_linear_k002"],
        nargs="+",
    )
    parser.add_argument("-p", "--pwakes", help="The partial wakes model", default=None)
    parser.add_argument("--ws", help="The wind speed", type=float, default=9.0)
    parser.add_argument("--wd", help="The wind direction", type=float, default=270.0)
    parser.add_argument("--ti", help="The TI value", type=float, default=0.08)
    parser.add_argument("--rho", help="The air density", type=float, default=1.225)
    parser.add_argument(
        "-d",
        "--min_dist",
        help="Minimal turbine distance in unit D",
        type=float,
        default=None,
    )
    parser.add_argument(
        "-A", "--opt_algo", help="The pymoo algorithm name", default="GA"
    )
    parser.add_argument(
        "-P", "--n_pop", help="The population size", type=int, default=50
    )
    parser.add_argument(
        "-G", "--n_gen", help="The nmber of generations", type=int, default=300
    )
    parser.add_argument(
        "-nop", "--no_pop", help="Switch off vectorization", action="store_true"
    )
    parser.add_argument("-e", "--engine", help="The engine", default="process")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    args = parser.parse_args()

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    boundary = foxes.utils.geom2d.Circle([0.0, 0.0], 1000.0)

    farm = foxes.WindFarm(boundary=boundary)
    foxes.input.farm_layout.add_row(
        farm=farm,
        xy_base=np.zeros(2),
        xy_step=np.array([50.0, 0.0]),
        n_turbines=args.n_t,
        turbine_models=["kTI_02", ttype.name],
    )
    states = foxes.input.states.SingleStateStates(
        ws=args.ws, wd=args.wd, ti=args.ti, rho=args.rho
    )

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame="rotor_wd",
        partial_wakes=args.pwakes,
        mbook=mbook,
        verbosity=0,
    )

    problem = FarmLayoutOptProblem("layout_opt", algo)
    problem.add_objective(MaxFarmPower(problem))
    problem.add_constraint(FarmBoundaryConstraint(problem, disc_inside=True))
    if args.min_dist is not None:
        problem.add_constraint(
            MinDistConstraint(problem, min_dist=args.min_dist, min_dist_unit="D")
        )
    problem.initialize()

    solver = Optimizer_pymoo(
        problem,
        problem_pars=dict(
            vectorize=not args.no_pop,
        ),
        algo_pars=dict(
            type=args.opt_algo,
            pop_size=args.n_pop,
            seed=None,
        ),
        setup_pars=dict(),
        term_pars=dict(
            type="default",
            n_max_gen=args.n_gen,
            ftol=1e-6,
            xtol=1e-6,
        ),
    )
    solver.initialize()
    solver.print_info()

    ax = foxes.output.FarmLayoutOutput(farm).get_figure()
    plt.show()
    plt.close(ax.get_figure())

    engine = foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
        verbosity=0,
    )

    with engine:
        results = solver.solve()
        solver.finalize(results)

        print()
        print(results)

        p_min = np.array([-1100.0, -1100.0])
        p_max = np.array([1100.0, 1100.0])
        o = foxes.output.FlowPlots2D(algo, results.problem_results)
        plot_data = o.get_mean_data_xy(
            "WS",
            resolution=20,
            xmin=p_min[0],
            xmax=p_max[0],
            ymin=p_min[1],
            ymax=p_max[1],
        )

    fig, axs = plt.subplots(1, 2, figsize=(12, 8))

    foxes.output.FarmLayoutOutput(farm).get_figure(fig=fig, ax=axs[0])

    o = foxes.output.FlowPlots2D(algo, results.problem_results)
    fig = o.get_mean_fig_xy(plot_data, fig=fig, ax=axs[1])
    dpars = dict(alpha=0.6, zorder=10, p_min=p_min, p_max=p_max)
    farm.boundary.add_to_figure(axs[1], fill_mode="outside_white", pars_distance=dpars)

    plt.show()
    plt.close(fig)



import numpy as np
import argparse
import matplotlib.pyplot as plt
from iwopy import LocalFD
from iwopy.optimizers import GG

import foxes
import foxes.variables as FV
from foxes_opt.problems.layout import FarmLayoutOptProblem
from foxes_opt.constraints import FarmBoundaryConstraint, MinDistConstraint
from foxes_opt.objectives import MaxFarmPower

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-nt", "--n_t", help="The number of turbines", type=int, default=10
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument(
        "-s",
        "--states",
        help="The states input file (path or static)",
        default="wind_rose_bremen.csv",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Bastankhah025_linear_k002"],
        nargs="+",
    )
    parser.add_argument("-p", "--pwakes", help="The partial wakes model", default=None)
    parser.add_argument("--ti", help="The TI value", type=float, default=0.04)
    parser.add_argument("--rho", help="The air density", type=float, default=1.225)
    parser.add_argument(
        "-d",
        "--min_dist",
        help="Minimal turbine distance in unit D",
        type=float,
        default=None,
    )
    parser.add_argument(
        "-nop", "--no_pop", help="Switch off vectorization", action="store_true"
    )
    parser.add_argument(
        "-O",
        "--fd_order",
        help="Finite difference derivative order",
        type=int,
        default=1,
    )
    parser.add_argument("-e", "--engine", help="The engine", default="process")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    args = parser.parse_args()

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    boundary = (
        foxes.utils.geom2d.ClosedPolygon(
            np.array([[0, 0], [0, 1200], [1000, 800], [900, -200]], dtype=np.float64)
        )
        + foxes.utils.geom2d.ClosedPolygon(
            np.array([[500, 0], [500, 1500], [1000, 1500], [1000, 0]], dtype=np.float64)
        )
        - foxes.utils.geom2d.Circle([-100.0, -100.0], 700)
    )

    farm = foxes.WindFarm(boundary=boundary)
    foxes.input.farm_layout.add_row(
        farm=farm,
        xy_base=np.array([500.0, 500.0]),
        xy_step=np.array([50.0, 50.0]),
        n_turbines=args.n_t,
        turbine_models=[ttype.name],
    )

    states = foxes.input.states.StatesTable(
        data_source=args.states,
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
        var2col={FV.WS: "ws", FV.WD: "wd", FV.WEIGHT: "weight"},
        fixed_vars={FV.RHO: args.rho, FV.TI: args.ti},
    )

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame="rotor_wd",
        partial_wakes=args.pwakes,
        mbook=mbook,
        verbosity=0,
    )

    problem = FarmLayoutOptProblem("layout_opt", algo)
    problem.add_objective(MaxFarmPower(problem))
    problem.add_constraint(FarmBoundaryConstraint(problem))
    if args.min_dist is not None:
        problem.add_constraint(
            MinDistConstraint(problem, min_dist=args.min_dist, min_dist_unit="D")
        )
    gproblem = LocalFD(problem, deltas=0.1, fd_order=args.fd_order)
    gproblem.initialize()

    solver = GG(
        gproblem,
        step_max=100.0,
        step_min=0.1,
        f_tol=1e-4,
        vectorized=not args.no_pop,
    )
    solver.initialize()
    solver.print_info()

    ax = foxes.output.FarmLayoutOutput(farm).get_figure()
    plt.show()
    plt.close(ax.get_figure())

    o = foxes.output.StatesRosePlotOutput(states, point=[0.0, 0.0, 100.0])
    fig = o.get_figure(16, FV.AMB_WS, [0, 3.5, 6, 10, 15, 20])
    plt.show()
    plt.close()

    engine = foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
        verbosity=0,
    )

    with engine:
        results = solver.solve()
        solver.finalize(results)

        print()
        print(results)

        o = foxes.output.FlowPlots2D(algo, results.problem_results)
        p_min = np.array([-100.0, -350.0])
        p_max = np.array([1100.0, 1600.0])
        plot_data = o.get_mean_data_xy(
            "WS",
            resolution=20,
            xmin=p_min[0],
            xmax=p_max[0],
            ymin=p_min[1],
            ymax=p_max[1],
        )

    fig, axs = plt.subplots(1, 2, figsize=(12, 8))

    foxes.output.FarmLayoutOutput(farm).get_figure(fig=fig, ax=axs[0])

    fig = o.get_mean_fig_xy(plot_data, fig=fig, ax=axs[1])
    dpars = dict(alpha=0.6, zorder=10, p_min=p_min, p_max=p_max)
    farm.boundary.add_to_figure(axs[1], fill_mode="outside_white", pars_distance=dpars)

    plt.show()
    plt.close(fig)



"""
Abstract classes and core functionality for wind farm optimization.
"""

from .farm_opt_problem import FarmOptProblem as FarmOptProblem
from .farm_vars_problem import FarmVarsProblem as FarmVarsProblem
from .farm_objective import FarmObjective as FarmObjective
from .farm_constraint import FarmConstraint as FarmConstraint



"""
Purely geometric wind farm layout problems.
"""

from .geom_reggrid import GeomRegGrid as GeomRegGrid
from .geom_reggrids import GeomRegGrids as GeomRegGrids

from .objectives import OMaxN as OMaxN
from .objectives import OMinN as OMinN
from .objectives import OFixN as OFixN
from .objectives import MaxGridSpacing as MaxGridSpacing
from .objectives import MaxDensity as MaxDensity
from .objectives import MeMiMaDist as MeMiMaDist

from .constraints import Valid as Valid
from .constraints import Boundary as Boundary
from .constraints import MinDist as MinDist
from .constraints import CMinN as CMinN
from .constraints import CMaxN as CMaxN
from .constraints import CFixN as CFixN
from .constraints import CMinDensity as CMinDensity



from xarray import Dataset
from iwopy.core import SingleObjOptResults, MultiObjOptResults
from foxes.utils import write_nc

from .opt_output import OptOutput


class SingleObjResultsWriter(OptOutput):
    """
    Writes optimization results to files.

    :group: output

    """

    def __init__(self, opt_results, **kwargs):
        """
        Constructor

        Parameters
        ----------
        opt_results: iwopy.core.SingleObjOptResults
            The optimization results
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(**kwargs)
        self.results = opt_results

        if not isinstance(opt_results, SingleObjOptResults):
            raise TypeError(
                f"{type(self).__name__}: Require results of type 'SingleObjOptResults', got '{type(opt_results).__name__}'"
            )

    def get_dataset(self):
        """
        Translates results into dataset

        Returns
        -------
        ds: xarray.Dataset
            The results dataset

        """

        crds = {}
        dvars = {}
        if len(self.results.vnames_int):
            crds["variables_int"] = self.results.vnames_int
            dvars["vars_int"] = ("variables_int", self.results.vars_int)
        if len(self.results.vnames_float):
            crds["variables_float"] = self.results.vnames_float
            dvars["vars_float"] = ("variables_float", self.results.vars_float)
        if len(self.results.cnames):
            crds["constraints"] = self.results.cnames
            dvars["cons"] = ("constraints", self.results.cons)

        ds = Dataset(
            coords=crds,
            data_vars=dvars,
            attrs=dict(
                problem=self.results.pname,
                success=int(self.results.success),
                objective=self.results.onames[0],
                obj_value=self.results.objs,
            ),
        )

        return ds

    def write_nc(self, fname, verbosity=1, **kwargs):
        """
        Writes to NetCDF file

        Parameters
        ----------
        fname: str
            Name of the file to be written
        verbosity: int
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Parameters for foxes.utils.write_nc

        """
        ds = self.get_dataset()
        fpath = self.get_fpath(fname)
        write_nc(ds, fpath, verbosity=verbosity, **kwargs)


class MultiObjResultsWriter(OptOutput):
    """
    Writes optimization results to files.

    :group: output

    """

    def __init__(self, opt_results, **kwargs):
        """
        Constructor

        Parameters
        ----------
        opt_results: iwopy.core.MultiObjOptResults
            The optimization results
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(**kwargs)
        self.results = opt_results

        if not isinstance(opt_results, MultiObjOptResults):
            raise TypeError(
                f"{type(self).__name__}: Require results of type 'MultiObjOptResults', got '{type(opt_results).__name__}'"
            )

    def get_dataset(self):
        """
        Translates results into dataset

        Returns
        -------
        ds: xarray.Dataset
            The results dataset

        """

        crds = {}
        dvars = {}
        if len(self.results.vnames_int):
            crds["variables_int"] = self.results.vnames_int
            dvars["vars_int"] = (("pop", "variables_int"), self.results.vars_int)
        if len(self.results.vnames_float):
            crds["variables_float"] = self.results.vnames_float
            dvars["vars_float"] = (("pop", "variables_float"), self.results.vars_float)
        if len(self.results.onames):
            crds["objectives"] = self.results.onames
            dvars["objs"] = (("pop", "objectives"), self.results.objs)
        if len(self.results.cnames):
            crds["constraints"] = self.results.cnames
            dvars["cons"] = (("pop", "constraints"), self.results.cons)

        ds = Dataset(
            coords=crds,
            data_vars=dvars,
            attrs=dict(
                problem=self.results.pname,
                success=int(self.results.success),
            ),
        )

        return ds

    def write_nc(self, fname, verbosity=1, **kwargs):
        """
        Writes to NetCDF file

        Parameters
        ----------
        fname: str
            Name of the file to be written
        verbosity: int
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Parameters for foxes.utils.write_nc

        """
        ds = self.get_dataset()
        fpath = self.get_fpath(fname)
        write_nc(ds, fpath, verbosity=verbosity, **kwargs)



import argparse
from pathlib import Path

from foxes.utils import Dict

from .dict import run_dict


def foxes_opt_yaml():
    """
    Command line tool for running foxes from yaml file input.

    Examples
    --------
    >>> foxes_opt_yaml input.yaml

    :group: input.yaml

    """

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "yml_file",
        help="The input yaml file",
    )
    parser.add_argument("-o", "--output_dir", help="The output directory", default=".")
    parser.add_argument("-r", "--rotor", help="The rotor model", default=None)
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default=None, nargs="+"
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=None,
        nargs="+",
    )
    parser.add_argument("-f", "--frame", help="The wake frame", default=None)
    parser.add_argument("-e", "--engine", help="The engine", default=None)
    parser.add_argument(
        "-n", "--n_procs", help="The number of processes", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-it", "--iterative", help="Use iterative algorithm", action="store_true"
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        help="The verbosity level, 0 = silent",
        type=int,
        default=None,
    )
    args = parser.parse_args()

    v = 1 if args.verbosity is None else args.verbosity
    fpath = Path(args.yml_file)
    idata = Dict.from_yaml(fpath, verbosity=v)

    if (
        args.engine is not None
        or args.n_procs is not None
        or args.chunksize_states is not None
        or args.chunksize_points is not None
    ):
        epars = dict(
            engine_type=args.engine,
            n_procs=args.n_procs,
            chunk_size_states=args.chunksize_states,
            chunk_size_points=args.chunksize_points,
            verbosity=v,
        )
    else:
        epars = None

    run_dict(
        idata,
        rotor_model=args.rotor,
        partial_wakes=args.pwakes,
        wake_models=args.wakes,
        wake_frame=args.frame,
        engine_pars=epars,
        iterative=args.iterative,
        work_dir=fpath.parent,
        output_dir=args.output_dir,
        verbosity=args.verbosity,
    )



from iwopy import LocalFD
from iwopy.core import Optimizer
from foxes.input.yaml import read_dict as foxes_read_dict
from foxes.input.yaml import run_outputs as foxes_run_output
from foxes.utils import Dict

from foxes_opt.core import FarmOptProblem, FarmObjective, FarmConstraint


def read_dict(idict, *args, verbosity=None, **kwargs):
    """
    Read dictionary input into foxes objects

    Parameters
    ----------
    idict: foxes.utils.Dict
        The input parameter dictionary
    args: tuple, optional
        Additional parameters for foxes.input.run_dict
    verbosity: int, optional
        Force a verbosity level, 0 = silent, overrules
        settings from idict
    kwargs: dict, optional
        Additional parameters for foxes.input.run_dict

    Returns
    -------
    algo: foxes.core.Algorithm
        The algorithm
    engine: foxes.core.Engine
        The engine, or None if not set
    optimizer: iwopy.core.Optimizer
        The optimization problem solver

    :group: input.yaml

    """

    def _print(*args, level=1, **kwargs):
        if verbosity is None or verbosity >= level:
            print(*args, **kwargs)

    # extract data:
    jdict = idict.pop_item("optimization")

    # read base components:
    algo, engine = foxes_read_dict(idict, *args, verbosity=verbosity, **kwargs)
    if engine is not None:
        engine.verbosity = 0

    # create problem:
    _print("Creating problem")
    pdict = jdict.get_item("problem")
    ldict = pdict.pop("local_fd", None)
    odicts = [
        Dict(o, _name=f"{pdict.name}.objective{i}")
        for i, o in enumerate(pdict.pop_item("objectives"))
    ]
    cdicts = pdict.pop("constraints", [])
    cdicts = [
        Dict(c, _name=f"{pdict.name}.constraint{i}") for i, c in enumerate(cdicts)
    ]
    flist = [
        Dict(f, _name=f"{pdict.name}.function{i}")
        for i, f in enumerate(pdict.pop("functions", []))
    ]
    problem = FarmOptProblem.new(algo=algo, **pdict)
    for fdict in flist:
        fname = fdict.pop_item("name")
        _print(f"  - {fname}")
        f = getattr(problem, fname)
        f(**fdict)
    for odict in odicts:
        _print(f"  Adding objective: {odict.get_item('objective_type')}")
        o = FarmObjective.new(problem=problem, **odict)
        problem.add_objective(o)
    for cdict in cdicts:
        _print(f"  Adding constraint: {odict.get_item('constraint_type')}")
        c = FarmConstraint.new(problem=problem, **cdict)
        problem.add_constraint(c)
    if ldict is not None:
        _print("Adding local finite differences")
        problem0 = problem
        problem = LocalFD(problem0, **ldict)
    problem.initialize()

    # create solver:
    _print("Creating optimizer")
    sdict = jdict.get_item("optimizer")
    optimizer = Optimizer.new(problem=problem, **sdict)
    optimizer.initialize()

    return algo, engine, optimizer


def run_outputs(
    idict,
    algo=None,
    farm_results=None,
    opt_results=None,
    extra_sig={},
    **kwargs,
):
    """
    Run outputs from dict.

    Parameters
    ----------
    idict: foxes.utils.Dict
        The input parameter dictionary
    algo: foxes.core.Algorithm, optional
        The algorithm
    farm_results: xarray.Dataset, optional
        The farm results
    opt_results: iwopy.core.SingleObjOptResults or iwopy.core.MultiObjOptResults, optional
        The optimization results
    extra_sig: dict
        Extra function signature check, sets
        arguments (key) with data (value)
    kwargs: dict, optional
        Additional parameters for foxes_run_output

    Returns
    -------
    outputs: list of tuple
        For each output enty, a tuple (dict, results),
        where results is a tuple that represents one
        entry per function call

    :group: input.yaml

    """
    extra_sig.update({"opt_results": opt_results})

    out = foxes_run_output(
        idict,
        algo,
        farm_results=farm_results,
        point_results=None,
        extra_sig=extra_sig,
        **kwargs,
    )

    return out


def run_dict(idict, *args, extra_sig={}, nofig=False, verbosity=None, **kwargs):
    """
    Run from a dictionary type parameter file.

    Parameters
    ----------
    idict: foxes.utils.Dict
        The input parameter dictionary
    args: tuple, optional
        Additional parameters for foxes.input.run_dict
    extra_sig: dict
        Extra function signature check, sets
        arguments (key) with data (value)
    nofig: bool
        Do not show figures, overrules settings from idict
    verbosity: int, optional
        Force a verbosity level, 0 = silent, overrules
        settings from idict
    kwargs: dict, optional
        Additional parameters for foxes.input.run_dict

    Returns
    -------
    opt_results: iwopy.core.SingleObjOptResults or iwopy.core.MultiObjOptResults
        The optimization results
    outputs: list of tuple
        For each output enty, a tuple (dict, results),
        where results is a tuple that represents one
        entry per function call

    """

    def _print(*args, level=1, **kwargs):
        if verbosity is None or verbosity >= level:
            print(*args, **kwargs)

    # read components:
    algo, engine, optimizer = read_dict(idict, *args, verbosity=verbosity, **kwargs)

    if verbosity is None or verbosity >= 0:
        optimizer.print_info()

    # run optimizer:
    rdict = idict.get_item("solve", Dict(_name=idict.name + ".solve"))
    results_storage = None
    if rdict.pop_item("run", True):
        _print("Running optimizer")
        with engine:
            opt_results = optimizer.solve(**rdict)
            optimizer.finalize(opt_results)
            farm_results = opt_results.problem_results

            # run outputs with engine:
            out_w, results_storage = run_outputs(
                idict,
                algo,
                farm_results,
                opt_results,
                extra_sig=extra_sig,
                with_engine=True,
                nofig=nofig,
                results_storage=results_storage,
                ret_results_storage=True,
                verbosity=verbosity,
            )
            out_w = list(out_w)

    else:
        opt_results = None
        farm_results = None

    print()
    print(opt_results)
    print()

    # run outputs w/o engine:
    out_wo = list(
        run_outputs(
            idict,
            algo,
            farm_results,
            opt_results,
            extra_sig=extra_sig,
            with_engine=False,
            nofig=nofig,
            results_storage=results_storage,
            ret_results_storage=True,
            verbosity=verbosity,
        ),
    )

    # combine outputs:
    out = tuple(a if a is not None else b for a, b in zip(out_w, out_wo))

    return opt_results, out



"""
Wind farm optimization constraints.
"""

from .area_geometry import AreaGeometryConstraint as AreaGeometryConstraint
from .area_geometry import FarmBoundaryConstraint as FarmBoundaryConstraint

from .min_dist import MinDistConstraint as MinDistConstraint



import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.distance import cdist
from iwopy import Problem

from foxes.config import config


class GeomLayoutGridded(Problem):
    """
    A layout within a boundary geometry, purely
    defined by geometrical optimization (no wakes),
    on a fixed background point grid along (x, y)
    directions.

    This optimization problem does not involve
    wind farms. Note that this is an integer problem.

    Attributes
    ----------
    boundary: foxes.utils.geom2d.AreaGeometry
        The boundary geometry
    n_turbines: int
        The number of turbines in the layout
    grid_spacing: float
        The background grid spacing
    min_dist: float
        The minimal distance between points
    D: float
        The diameter of circle fully within boundary

    :group: opt.problems.layout.geom_layouts

    """

    def __init__(
        self,
        boundary,
        n_turbines,
        grid_spacing,
        min_dist=None,
        D=None,
    ):
        """
        Constructor.

        Parameters
        ----------
        boundary: foxes.utils.geom2d.AreaGeometry
            The boundary geometry
        n_turbines: int
            The number of turbines in the layout
        grid_spacing: float
            The background grid spacing
        min_dist: float, optional
            The minimal distance between points
        D: float, optional
            The diameter of circle fully within boundary

        """
        super().__init__(name="geom_reg_grids")

        self.boundary = boundary
        self.n_turbines = n_turbines
        self.grid_spacing = grid_spacing
        self.D = D
        self.min_dist = min_dist

        self._I = [f"i{i}" for i in range(self.n_turbines)]

    def initialize(self, verbosity=1):
        """
        Initialize the object.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().initialize(verbosity)

        pmin = self.boundary.p_min()
        pmax = self.boundary.p_max() + self.grid_spacing
        self._pts = np.stack(
            np.meshgrid(
                np.arange(pmin[0], pmax[0], self.grid_spacing),
                np.arange(pmin[1], pmax[1], self.grid_spacing),
                indexing="ij",
            ),
            axis=-1,
        )
        nx, ny = self._pts.shape[:2]
        self._pts = self._pts.reshape(nx * ny, 2)

        if self.D is None:
            valid = self.boundary.points_inside(self._pts)
        else:
            valid = self.boundary.points_inside(self._pts) & (
                self.boundary.points_distance(self._pts) >= self.D / 2
            )
        self._pts = self._pts[valid]
        self._N = len(self._pts)

        if verbosity > 0:
            print(f"Problem '{self.name}': n_bgd_pts = {self._N}")

        if self._N < self.n_turbines:
            raise ValueError(
                f"Problem '{self.name}': Background grid only provides {self._N} points for {self.n_turbines} turbines"
            )

        self.apply_individual(self.initial_values_int(), self.initial_values_float())

    def var_names_int(self):
        """
        The names of int variables.

        Returns
        -------
        names: list of str
            The names of the int variables

        """
        return self._I

    def initial_values_int(self):
        """
        The initial values of the int variables.

        Returns
        -------
        values: numpy.ndarray
            Initial int values, shape: (n_vars_int,)

        """
        return np.arange(self.n_turbines, dtype=config.dtype_int)

    def min_values_int(self):
        """
        The minimal values of the int variables.

        Returns
        -------
        values: numpy.ndarray
            Minimal int values, shape: (n_vars_int,)

        """
        return np.zeros(self.n_turbines, dtype=config.dtype_int)

    def max_values_int(self):
        """
        The maximal values of the int variables.

        Returns
        -------
        values: numpy.ndarray
            Maximal int values, shape: (n_vars_int,)

        """
        return np.full(self.n_turbines, self._N - 1, dtype=config.dtype_int)

    def apply_individual(self, vars_int, vars_float):
        """
        Apply new variables to the problem.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float: np.array
            The float variable values, shape: (n_vars_float,)

        Returns
        -------
        problem_results: Any
            The results of the variable application
            to the problem

        """
        xy = self._pts[vars_int.astype(config.dtype_int)]
        __, ui = np.unique(vars_int, return_index=True)
        valid = np.zeros(self.n_turbines, dtype=bool)
        valid[ui] = True
        return xy, valid

    def apply_population(self, vars_int, vars_float):
        """
        Apply new variables to the problem,
        for a whole population.

        Parameters
        ----------
        vars_int: np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float: np.array
            The float variable values, shape: (n_pop, n_vars_float)

        Returns
        -------
        problem_results: Any
            The results of the variable application
            to the problem

        """
        n_pop = vars_int.shape[0]

        vint = vars_int.reshape(n_pop * self.n_turbines).astype(config.dtype_int)
        xy = self._pts[vint, :].reshape(n_pop, self.n_turbines, 2)

        valid = np.zeros((n_pop, self.n_turbines), dtype=bool)
        for pi in range(n_pop):
            __, ui = np.unique(vars_int[pi], return_index=True)
            valid[pi, ui] = True

        return xy, valid

    def get_fig(
        self, xy=None, valid=None, ax=None, title=None, true_circle=True, **bargs
    ):
        """
        Return plotly figure axis.

        Parameters
        ----------
        xy: numpy.ndarary, optional
            The xy coordinate array, shape: (n_points, 2)
        valid: numpy.ndarray, optional
            Boolean array of validity, shape: (n_points,)
        ax: pyplot.Axis, optional
            The figure axis
        title: str, optional
            The figure title
        true_circle: bool
            Draw points as circles with diameter self.D
        bars: dict, optional
            The boundary plot arguments

        Returns
        -------
        ax: pyplot.Axis
            The figure axis

        """
        if ax is None:
            __, ax = plt.subplots()

        hbargs = {"fill_mode": "inside_lightgray"}
        hbargs.update(bargs)
        self.boundary.add_to_figure(ax, **hbargs)

        if xy is not None:
            if valid is not None:
                xy = xy[valid]
            if not true_circle or self.D is None:
                ax.scatter(xy[:, 0], xy[:, 1], color="orange")
            else:
                for x, y in xy:
                    ax.add_patch(
                        plt.Circle((x, y), self.D / 2, color="blue", fill=True)
                    )

        ax.set_aspect("equal", adjustable="box")
        ax.set_xlabel("x [m]")
        ax.set_ylabel("y [m]")

        if title is None:
            if xy is None:
                title = "Optimization area"
            else:
                lxy = len(xy) if xy is not None else 0
                dists = cdist(xy, xy)
                np.fill_diagonal(dists, 1e20)
                title = f"N = {lxy}, min_dist = {np.min(dists):.1f} m"
        ax.set_title(title)

        return ax



import numpy as np
import pandas as pd

from foxes_opt.core import FarmVarsProblem
from foxes.models.turbine_models import SetFarmVars
from foxes.config import config


class OptFarmVars(FarmVarsProblem):
    """
    Optimize a selection of farm variables.

    :group: opt.problems

    """

    def __init__(self, *args, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Arguments for `FarmVarsProblem`
        kwargs: dict, optional
            Keyword arguments for `FarmVarsProblem`

        """
        super().__init__(*args, **kwargs)
        self._vars = None

    def add_var(
        self,
        variable,
        typ,
        init,
        min,
        max,
        level="uniform",
        sel=None,
        pre_rotor=False,
        model_key=None,
    ):
        """
        Add a variable.

        Parameters
        ----------
        variable: str
            The foxes farm variable name
        typ: type
            The variable type, either float or int
        init: float or int
            The initial value
        min: float or int
            The min value
        max: float or int
            The max value
        level: str
            Choices: uniform, state, turbine, state-turbine
        sel: numpy.ndarray, optional
            States/turbines/state-turbine selection,
            depending on the level
        pre_rotor: bool
            Apply this variable before rotor model
        model_key: str, optional
            Creates sub-model which can then be placed in the
            turbine model list. Repeated keys are added to the
            same turbine model

        """
        if typ == "float":
            typ = float
        elif typ == "int":
            typ = int
        elif typ is not float and typ is not int:
            raise TypeError(
                f"Problem '{self.name}': Expecting float or int, got type '{type(typ).__name__}'"
            )

        mname = self.name if model_key is None else model_key
        if mname in self.algo.mbook.turbine_models:
            m = self.algo.mbook.turbine_models[mname]
            if not isinstance(m, SetFarmVars):
                raise KeyError(
                    f"Problem '{self.name}': Turbine model entry '{mname}' already exists in model book, and is not of type SetFarmVars"
                )
            elif m.pre_rotor != pre_rotor:
                raise ValueError(
                    f"Problem '{self.name}': Turbine model entry '{mname}' exists in model book, and disagrees on pre_rotor = {pre_rotor}"
                )
        else:
            self.algo.mbook.turbine_models[mname] = SetFarmVars(pre_rotor=pre_rotor)

        if self._vars is None:
            i0 = 0
            i0i = 0
            i0f = 0
        else:
            if variable in self._vars["var"].tolist():
                raise ValueError(
                    f"Problem '{self.name}': Attempt to add variable '{variable}' twice"
                )
            i0 = len(self._vars.index)
            grps = self._vars.groupby("type")
            i0i = len(grps.get_group("int").index) if "int" in grps.groups.keys() else 0
            i0f = (
                len(grps.get_group("float").index)
                if "float" in grps.groups.keys()
                else 0
            )
            del grps

        if level == "uniform":
            hdata = pd.DataFrame(index=[i0])
            hdata.loc[i0, "name"] = variable
            hdata.loc[i0, "var"] = variable
            hdata.loc[i0, "type"] = "int" if typ is int else "float"
            hdata.loc[i0, "index"] = i0i if typ is int else i0f
            hdata.loc[i0, "level"] = level
            hdata.loc[i0, "state"] = -1
            hdata.loc[i0, "turbine"] = -1
            hdata.loc[i0, "sel_turbine"] = -1
            hdata.loc[i0, "init"] = np.array([init], dtype=config.dtype_double)
            hdata.loc[i0, "min"] = np.array([min], dtype=config.dtype_double)
            hdata.loc[i0, "max"] = np.array([max], dtype=config.dtype_double)
            hdata.loc[i0, "pre_rotor"] = pre_rotor
            hdata.loc[i0, "model_key"] = mname

        elif level == "state":
            if not self.algo.initialized:
                self.algo.initialize()

            states = np.arange(self.algo.n_states)
            if sel is not None:
                states = states[sel]
            inds = i0 + np.arange(len(states))
            tinds = inds - i0 + i0i if typ is int else inds - i0 + i0f

            hdata = pd.DataFrame(index=inds)
            hdata.loc[inds, "name"] = [
                f"{variable}_{i:05d}" for i in range(len(states))
            ]
            hdata.loc[inds, "var"] = variable
            hdata.loc[inds, "type"] = "int" if typ is int else "float"
            hdata.loc[inds, "index"] = tinds
            hdata.loc[inds, "level"] = level
            hdata.loc[inds, "state"] = states
            hdata.loc[inds, "turbine"] = -1
            hdata.loc[inds, "sel_turbine"] = -1

            for c, d in [("init", init), ("min", min), ("max", max)]:
                data = np.full(len(inds), np.nan, dtype=config.dtype_double)
                data[:] = d
                hdata.loc[inds, c] = data

            hdata.loc[inds, "pre_rotor"] = pre_rotor
            hdata.loc[inds, "model_key"] = mname

        elif level == "turbine":
            if sel is None:
                turbines = self.sel_turbines
            else:
                turbines = np.arange(self.algo.n_turbines)[sel]
            inds = i0 + np.arange(len(turbines))
            tinds = inds - i0 + i0i if typ is int else inds - i0 + i0f

            hdata = pd.DataFrame(index=inds)
            hdata.loc[inds, "name"] = [
                f"{variable}_{i:04d}" for i in range(len(turbines))
            ]
            hdata.loc[inds, "var"] = variable
            hdata.loc[inds, "type"] = "int" if typ is int else "float"
            hdata.loc[inds, "index"] = tinds
            hdata.loc[inds, "level"] = level
            hdata.loc[inds, "state"] = -1
            hdata.loc[inds, "turbine"] = turbines
            hdata.loc[inds, "sel_turbine"] = [
                self.sel_turbines.index(ti) for ti in turbines
            ]

            for c, d in [("init", init), ("min", min), ("max", max)]:
                data = np.full(len(inds), np.nan, dtype=config.dtype_double)
                data[:] = d
                hdata.loc[inds, c] = data

            hdata.loc[inds, "pre_rotor"] = pre_rotor
            hdata.loc[inds, "model_key"] = mname

        elif level == "state-turbine":
            if not self.algo.initialized:
                self.algo.initialize()

            n_states = self.algo.n_states
            n_turbines = self.algo.n_turbines
            if sel is None:
                sel = np.zeros((n_states, n_turbines), dtype=bool)
                sel[:, self.sel_turbines] = True
            else:
                sel = np.array(sel, dtype=bool)
            st = np.arange(n_states * n_turbines).reshape(n_states, n_turbines)[sel]
            whr = np.where(sel)
            n_inds = len(st)
            inds = i0 + np.arange(n_inds)
            tinds = inds - i0 + i0i if typ is int else inds - i0 + i0f

            hdata = pd.DataFrame(index=inds)
            hdata.loc[inds, "name"] = [
                f"{variable}_{whr[0][i]:05d}_{whr[1][i]:04d}" for i in range(len(st))
            ]
            hdata.loc[inds, "var"] = variable
            hdata.loc[inds, "type"] = "int" if typ is int else "float"
            hdata.loc[inds, "index"] = tinds
            hdata.loc[inds, "level"] = level
            hdata.loc[inds, "state"] = whr[0]
            hdata.loc[inds, "turbine"] = whr[1]
            hdata.loc[inds, "sel_turbine"] = [
                self.sel_turbines.index(ti) for ti in whr[1]
            ]

            for c, d in [("init", init), ("min", min), ("max", max)]:
                data = np.full(n_inds, np.nan, dtype=config.dtype_double)
                if isinstance(d, np.ndarray) and len(d.shape) > 1:
                    data[:] = d[sel]
                else:
                    data[:] = d
                hdata.loc[inds, c] = data

            hdata.loc[inds, "pre_rotor"] = pre_rotor
            hdata.loc[inds, "model_key"] = mname

        else:
            raise ValueError(
                f"Problem '{self.name}': Unknown level '{level}'. Choices: uniform, state, turbine, state-turbine"
            )

        if self._vars is None:
            self._vars = hdata
        else:
            self._vars = pd.concat([self._vars, hdata], axis=0)

        icols = ["index", "state", "turbine", "sel_turbine"]
        for c in icols:
            self._vars[c] = self._vars[c].astype(config.dtype_int)

    def initialize(self, verbosity=1, **kwargs):
        """
        Initialize the object.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Additional parameters for super class init

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        if verbosity > 0:
            print(f"Problem '{self.name}': Optimization variable list")
            print()
            print(self._vars)
            print()

        prev = {}
        postv = {}
        for (mname, pre), g in self._vars.groupby(["model_key", "pre_rotor"]):
            if (pre and mname in postv) or (not pre and mname in prev):
                raise ValueError(
                    f"Problem '{self.name}': Model '{mname}' reveived both pre_rotor and non-pre_rotor variables"
                )
            tg = prev if pre else postv
            if mname not in tg:
                tg[mname] = set(g["var"].tolist())
            else:
                tg[mname] = tg[mname].update(g["var"].tolist())

        super().initialize(
            pre_rotor_vars={mname: list(vrs) for mname, vrs in prev.items()},
            post_rotor_vars={mname: list(vrs) for mname, vrs in postv.items()},
            verbosity=verbosity,
            **kwargs,
        )

    def var_names_int(self):
        """
        The names of int variables.

        Returns
        -------
        names: list of str
            The names of the int variables

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "int" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("int")["name"].tolist()

    def initial_values_int(self):
        """
        The initial values of the int variables.

        Returns
        -------
        values: numpy.ndarray
            Initial int values, shape: (n_vars_int,)

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "int" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("int")["init"].to_numpy(config.dtype_int)

    def min_values_int(self):
        """
        The minimal values of the integer variables.

        Use -self.INT_INF for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Minimal int values, shape: (n_vars_int,)

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "int" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("int")["min"].to_numpy(config.dtype_int)

    def max_values_int(self):
        """
        The maximal values of the integer variables.

        Use self.INT_INF for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Maximal int values, shape: (n_vars_int,)

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "int" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("int")["max"].to_numpy(config.dtype_int)

    def var_names_float(self):
        """
        The names of float variables.

        Returns
        -------
        names: list of str
            The names of the float variables

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "float" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("float")["name"].tolist()

    def initial_values_float(self):
        """
        The initial values of the float variables.

        Returns
        -------
        values: numpy.ndarray
            Initial float values, shape: (n_vars_float,)

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "float" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("float")["init"].to_numpy(config.dtype_double)

    def min_values_float(self):
        """
        The minimal values of the float variables.

        Use -numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Minimal float values, shape: (n_vars_float,)

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "float" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("float")["min"].to_numpy(config.dtype_double)

    def max_values_float(self):
        """
        The maximal values of the float variables.

        Use numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Maximal float values, shape: (n_vars_float,)

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "float" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("float")["max"].to_numpy(config.dtype_double)

    def opt2farm_vars_individual(self, vars_int, vars_float):
        """
        Translates optimization variables to farm variables

        Parameters
        ----------
        vars_int: numpy.ndarray
            The integer optimization variable values,
            shape: (n_vars_int,)
        vars_float: numpy.ndarray
            The float optimization variable values,
            shape: (n_vars_float,)

        Returns
        -------
        farm_vars: dict
            The foxes farm variables. Key: var name,
            value: numpy.ndarray with values, shape:
            (n_states, n_sel_turbines)

        """
        n_states = self.algo.n_states
        n_sturb = self.n_sel_turbines

        farm_vars = {}
        grps = self._vars.groupby(["type", "var", "level"])
        for (typ, var, level), g in grps:
            src = vars_int if typ == "int" else vars_float
            i0 = g.index[0]
            i1 = g.index[-1]
            data = src[np.s_[i0 : i1 + 1]]

            if level == "uniform":
                farm_vars[var] = np.full(
                    (n_states, n_sturb), data[0], dtype=config.dtype_double
                )

            elif level == "state":
                farm_vars[var] = np.full(
                    (n_states, n_sturb), np.nan, dtype=config.dtype_double
                )
                if np.all(g["state"] == np.arange(n_states)):
                    farm_vars[var][:] = data[:, None]
                else:
                    farm_vars[var][g["state"]] = data[:, None]

            elif level == "turbine":
                farm_vars[var] = np.full(
                    (n_states, n_sturb), np.nan, dtype=config.dtype_double
                )
                if np.all(g["sel_turbine"] == np.arange(n_sturb)):
                    farm_vars[var][:] = data[None, :]
                else:
                    farm_vars[var][:, g["sel_turbine"]] = data[None, :]

            elif level == "state-turbine":
                farm_vars[var] = np.full(
                    (n_states, n_sturb), np.nan, dtype=config.dtype_double
                )
                farm_vars[var][g["state"], g["sel_turbine"]] = data

            else:
                raise ValueError(
                    f"Problem '{self.name}': Unknown level '{level}' encountered for variable '{var}'. Valid choices: uniform, state, turbine, state-turbine"
                )

        return farm_vars

    def opt2farm_vars_population(self, vars_int, vars_float, n_states):
        """
        Translates optimization variables to farm variables

        Parameters
        ----------
        vars_int: numpy.ndarray
            The integer optimization variable values,
            shape: (n_pop, n_vars_int)
        vars_float: numpy.ndarray
            The float optimization variable values,
            shape: (n_pop, n_vars_float)
        n_states: int
            The number of original (non-pop) states

        Returns
        -------
        farm_vars: dict
            The foxes farm variables. Key: var name,
            value: numpy.ndarray with values, shape:
            (n_pop, n_states, n_sel_turbines)

        """
        n_pop = vars_float.shape[0]
        n_sturb = self.n_sel_turbines

        farm_vars = {}
        grps = self._vars.groupby(["type", "var", "level"])
        for (typ, var, level), g in grps:
            src = vars_int if typ == "int" else vars_float
            i0 = g.index[0]
            i1 = g.index[-1]
            data = src[:, np.s_[i0 : i1 + 1]]

            if level == "uniform":
                farm_vars[var] = np.full(
                    (n_pop, n_states, n_sturb), np.nan, dtype=config.dtype_double
                )
                farm_vars[var][:] = data[:, 0, None, None]

            elif level == "state":
                farm_vars[var] = np.full(
                    (n_pop, n_states, n_sturb), np.nan, dtype=config.dtype_double
                )
                if np.all(g["state"] == np.arange(n_states)):
                    farm_vars[var][:] = data[:, :, None]
                else:
                    farm_vars[var][:, g["state"]] = data[:, :, None]

            elif level == "turbine":
                farm_vars[var] = np.full(
                    (n_pop, n_states, n_sturb), np.nan, dtype=config.dtype_double
                )
                if np.all(g["sel_turbine"] == np.arange(n_sturb)):
                    farm_vars[var][:] = data[:, None, :]
                else:
                    farm_vars[var][:, :, g["sel_turbine"]] = data[:, None, :]

            elif level == "state-turbine":
                farm_vars[var] = np.full(
                    (n_pop, n_states, n_sturb), np.nan, dtype=config.dtype_double
                )
                farm_vars[var][:, g["state"], g["sel_turbine"]] = data

            else:
                raise ValueError(
                    f"Problem '{self.name}': Unknown level '{level}' encountered for variable '{var}'. Valid choices: uniform, state, turbine, state-turbine"
                )

        return farm_vars



"""
Wind farm layout optimization problems.
"""

from .farm_layout import FarmLayoutOptProblem as FarmLayoutOptProblem
from .regular_layout import RegularLayoutOptProblem as RegularLayoutOptProblem

from . import geom_layouts as geom_layouts



"""
Wind farm optimization.
"""

from .core import FarmOptProblem as FarmOptProblem
from .core import FarmObjective as FarmObjective
from .core import FarmConstraint as FarmConstraint

from . import input as input
from . import problems as problems
from . import constraints as constraints
from . import objectives as objectives
from . import output as output

import importlib
from pathlib import Path

try:
    tomllib = importlib.import_module("tomllib")
    source_location = Path(__file__).parent
    if (source_location.parent / "pyproject.toml").exists():
        with open(source_location.parent / "pyproject.toml", "rb") as f:
            __version__ = tomllib.load(f)["project"]["version"]
    else:
        __version__ = importlib.metadata.version(__package__ or __name__)
except ModuleNotFoundError:
    __version__ = importlib.metadata.version(__package__ or __name__)
